from abc import abstractmethod

from ..utils import Action, SetupOptions, StartOption
from ..model import Dealer, Player


class PlayerQuit(Exception):
    pass


class SaveGame(Exception):
    pass


class AbstractBlackjackView:  # pylint: disable=too-many-public-methods

    @abstractmethod
    def get_setup_options(self) -> SetupOptions:
        raise NotImplementedError

    @abstractmethod
    def get_bet(self, player: Player, *, save: bool = False) -> int:
        """Raises PlayerQuit if player indicates a quit.

        Raises SaveGame if player indicates to save the game."""
        raise NotImplementedError

    @abstractmethod
    def get_insurance_option(self, player: Player) -> bool:
        raise NotImplementedError

    @abstractmethod
    def get_action(self, actions: Action,
                   lowchips: Action = Action(0)) -> Action:
        raise NotImplementedError

    @abstractmethod
    def get_start_option(self) -> StartOption:
        raise NotImplementedError

    @abstractmethod
    def display_player_hand(self, player: Player, i: int) -> None:
        raise NotImplementedError

    @abstractmethod
    def display_dealer_hand(self, dealer: Dealer) -> None:
        raise NotImplementedError

    @abstractmethod
    def display_bet(self, player: Player) -> None:
        raise NotImplementedError

    @abstractmethod
    def display_blackjack(self, player: Player) -> None:
        raise NotImplementedError

    @abstractmethod
    def display_win(self, player: Player, i: int) -> None:
        raise NotImplementedError

    @abstractmethod
    def display_push(self, player: Player, i: int) -> None:
        raise NotImplementedError

    @abstractmethod
    def display_loss(self, player: Player, i: int) -> None:
        raise NotImplementedError

    @abstractmethod
    def display_bust(self, player: Player, i: int) -> None:
        raise NotImplementedError

    @abstractmethod
    def display_dealer_bust(self, dealer: Dealer) -> None:
        raise NotImplementedError

    @abstractmethod
    def display_dealer_blackjack(self, dealer: Dealer) -> None:
        raise NotImplementedError

    @abstractmethod
    def display_new_hand(self, *, new_shoe: bool = False) -> None:
        raise NotImplementedError

    @abstractmethod
    def display_start_of_actions(self) -> None:
        raise NotImplementedError

    @abstractmethod
    def display_end_of_actions(self) -> None:
        raise NotImplementedError

    @abstractmethod
    def display_player_elimination(self, player: Player) -> None:
        raise NotImplementedError

    @abstractmethod
    def display_exit(self) -> None:
        raise NotImplementedError

    @abstractmethod
    def get_save_filename(self) -> str:
        raise NotImplementedError

    @abstractmethod
    def confirm_overwrite(self) -> bool:
        raise NotImplementedError

    @abstractmethod
    def get_load_game(self) -> str | None:
        raise NotImplementedError



import os

from blessed import Terminal
from typing_extensions import Final, Literal, cast, overload, override

from ..model import Dealer, Hand, Player
from ..utils import Action, SetupOptions, SAVE_DIR, SAVE_EXT, StartOption
from ._abc import AbstractBlackjackView, PlayerQuit, SaveGame


DIVIDER_LENGTH: Final[int] = 10

ACTIONS_BY_LETTER: dict[str, Action] = {
    "h": Action.HIT,
    "s": Action.STAND,
    "p": Action.SPLIT,
    "d": Action.DOUBLE_DOWN,
    "u": Action.SURRENDER,
}

START_OPTIONS_BY_LETTER: dict[str, StartOption] = {
    "n": StartOption.NEW_GAME,
    "l": StartOption.LOAD_GAME,
}

LETTERS_BY_ACTION: dict[Action, str] = {}
for letter, action in ACTIONS_BY_LETTER.items():
    LETTERS_BY_ACTION[action] = letter

LETTERS_BY_START_OPTION: dict[StartOption, str] = {}
for letter, option in START_OPTIONS_BY_LETTER.items():
    LETTERS_BY_START_OPTION[option] = letter

ACTION_WORDS: dict[Action, str] = {
    Action.HIT: "[H]it",
    Action.STAND: "[S]tand",
    Action.SPLIT: "S[p]lit",
    Action.DOUBLE_DOWN: "[D]ouble down",
    Action.SURRENDER: "S[u]rrender",
}

START_OPTION_WORDS: dict[StartOption, str] = {
    StartOption.NEW_GAME: "[N]ew game",
    StartOption.LOAD_GAME: "[L]oad saved game",
}


class BlackjackView(AbstractBlackjackView):  # pylint: disable=too-many-public-methods

    def __init__(self) -> None:
        super().__init__()
        self.term = Terminal()

    @override
    def get_setup_options(self) -> SetupOptions:
        print("Welcome to PyBlackjack!")
        numplayers = cast(
            int | Literal["q"],
            self._get_int("Enter number of players [1-6] or [q]uickstart: ", 1,
                            6, "q"))
        if numplayers == "q":
            return SetupOptions()

        players: list[str] = []
        for n in range(numplayers):
            name: str = self._get_str(f"Enter name for player {n + 1}: ")
            players.append(name)
        chips: int = self._get_int("Enter starting chips: ", min=1)
        decks: int = self._get_int("Enter number of decks in the shoe: ", 1, 8)
        hit_soft_17: bool = self._get_yes_no("Should dealer hit on soft 17?")

        return SetupOptions(
            player_names=players,
            starting_chips=chips,
            decks=decks,
            hit_soft_17=hit_soft_17,
        )

    @override
    def get_bet(self, player: Player, *, save: bool = False) -> int:
        print(f"{player.name} has {player.chips} chips.")
        save_text = ", [s]ave," if save else ""
        bet = cast(
            int | Literal["q", "s"],
            self._get_int(f"Enter a bet{save_text} or [q]uit: ", 0, player.chips,
                            "qs"))
        if bet == "q":
            raise PlayerQuit
        if bet == "s":
            raise SaveGame
        return bet

    @override
    def get_insurance_option(self, player: Player) -> bool:
        return self._get_yes_no(f"{player.name}: Buy insurance?")

    @override
    def get_action(self, actions: Action,
                   lowchips: Action = Action(0)) -> Action:
        text = ", ".join(ACTION_WORDS[action] for action in actions) + "? "
        letters = "".join(LETTERS_BY_ACTION[action] for action in actions)
        lowchips_letters = "".join(LETTERS_BY_ACTION[action]
                                   for action in lowchips)
        choice = self._get_action(text,
                                  options=letters,
                                  lowchips=lowchips_letters)
        return ACTIONS_BY_LETTER[choice]

    @override
    def get_start_option(self) -> StartOption:
        choice = self._get_action("[N]ew game or [L]oad saved game? ",
                                  options="nl",
                                  lowchips="")
        return START_OPTIONS_BY_LETTER[choice]

    @override
    def display_player_hand(self, player: Player, i: int) -> None:
        assert i > 0
        hand_number: str = f" (hand {i})" if player.is_split() else ""
        hand: Hand = player.hands[i - 1]
        card_text: str = " ".join(str(card) for card in hand)
        print(
            f"{player.name}{hand_number}: {card_text} ({hand.total()})"
        )

    @override
    def display_dealer_hand(self, dealer: Dealer) -> None:
        print("Dealer: ", end="")
        cards = [str(card) for card in dealer]
        if dealer.downcard:
            cards[0] = "[]"
            print(" ".join(cards))
        else:
            print(f"{' '.join(cards)} ({dealer.total()})")

    @override
    def display_bet(self, player: Player) -> None:
        pass

    @override
    def display_blackjack(self, player: Player) -> None:
        print(f"{player.name} has blackjack!")

    def _hand_index_message(self, player: Player, i: int, /) -> str:
        assert i > 0
        return f" on hand {i}" if player.is_split() else ""

    @override
    def display_win(self, player: Player, i: int) -> None:
        msg = self._hand_index_message(player, i)
        print(f"{player.name} wins{msg}!")

    @override
    def display_push(self, player: Player, i: int) -> None:
        msg = self._hand_index_message(player, i)
        print(f"{player.name} pushes{msg}.")

    @override
    def display_loss(self, player: Player, i: int) -> None:
        msg = self._hand_index_message(player, i)
        print(f"{player.name} loses{msg}.")

    @override
    def display_bust(self, player: Player, i: int) -> None:
        msg = self._hand_index_message(player, i)
        print(f"{player.name} busted{msg}!")

    @override
    def display_dealer_bust(self, dealer: Dealer) -> None:
        pass

    @override
    def display_dealer_blackjack(self, dealer: Dealer) -> None:
        print("Dealer has blackjack!")

    @override
    def display_new_hand(self, *, new_shoe: bool = False) -> None:
        print("=" * DIVIDER_LENGTH)
        if new_shoe:
            print("New shoe in play!")
            print("=" * DIVIDER_LENGTH)

    @override
    def display_start_of_actions(self) -> None:
        print("-" * DIVIDER_LENGTH)

    @override
    def display_end_of_actions(self) -> None:
        print("-" * DIVIDER_LENGTH)

    @override
    def display_player_elimination(self, player: Player) -> None:
        print(f"{player.name} is out of chips and is eliminated.")

    @override
    def display_exit(self) -> None:
        print("Thanks for playing!")

    @override
    def get_save_filename(self) -> str:
        return self._get_str("Enter filename: ")

    @override
    def confirm_overwrite(self) -> bool:
        return self._get_yes_no("File exists. Overwrite?")

    @override
    def get_load_game(self) -> str | None:
        save_files: list[str] = []
        for _, _, files in os.walk(SAVE_DIR):
            for filename in files:
                if not filename.endswith(SAVE_EXT):
                    continue
                save_files.append((filename))
        if len(save_files) == 0:
            print("No saved games found. Starting a new game instead.")
            return None
        save_files.sort()
        for n, filename in enumerate(save_files, start=1):
            print(f"{n:>3}: {filename.removesuffix(SAVE_EXT)}")
        file_num = cast(
            int | Literal["n"],
            self._get_int("Select file by number or [n]ew game: ",
                          min=1,
                          max=len(save_files),
                          alt="n"))
        if file_num == "n":
            return None
        return save_files[file_num - 1]

    def _error(self, message: str, /) -> None:
        print(f"Error: {message} Try again.")

    def _get_str(self, prompt: str, /) -> str:
        while True:
            s: str = input(prompt).strip()
            if s:
                return s
            else:
                self._error("Input required.")

    def _get_action(self, prompt: str, /, options: str, lowchips: str) -> str:
        while True:
            print(prompt, end="", flush=True)
            with self.term.cbreak():
                s: str = self.term.inkey().lower()
            print()
            if s in options:
                return s
            elif s in lowchips:
                self._error("Insufficient chips for that action.")
            else:
                self._error("Invalid input.")

    def _get_yes_no(self, prompt: str, /) -> bool:
        while True:
            s: str = self._get_action(f"{prompt} [y/n]: ", "yn", "")
            return s == "y"

    @overload
    def _get_int(self,
                 prompt: str,
                 /,
                 min: int | None = None,
                 max: int | None = None,
                 alt: None = None) -> int:
        ...

    @overload
    def _get_int(self,
                 prompt: str,
                 /,
                 min: int | None = None,
                 max: int | None = None,
                 alt: str = "") -> int | str:
        ...

    def _get_int(self,
                 prompt: str,
                 /,
                 min: int | None = None,
                 max: int | None = None,
                 alt: str | None = None) -> int | str:
        while True:
            s: str = self._get_str(prompt).lower()
            if alt and s.lower() in alt.lower():
                return s.lower()
            try:
                z: int = int(s)
            except ValueError:
                self._error("Invalid integer.")
                continue
            if min is not None and z < min:
                self._error(f"Integer must be at least {min}.")
                continue
            if max is not None and max < z:
                self._error(f"Integer must be at most {max}.")
                continue
            return z



from __future__ import annotations

from dataclasses import dataclass
from enum import Flag, auto
from pathlib import Path
import sys

from typing_extensions import Final, Generator, Never, Sequence

SAVE_DIR: Final[Path] = Path.home() / ".pyblackjack"
SAVE_EXT: Final[str] = ".pybj"


class Action(Flag):
    HIT = auto()
    STAND = auto()
    SPLIT = auto()
    DOUBLE_DOWN = auto()
    SURRENDER = auto()

    if sys.version_info < (3, 11):
        def __iter__(self) -> Generator[Action]:
            for action in Action:
                if action in self:
                    yield action


class StartOption(Flag):
    NEW_GAME = auto()
    LOAD_GAME = auto()

    if sys.version_info < (3, 11):
        def __iter__(self) -> Generator[Action]:
            for action in Action:
                if action in self:
                    yield action


@dataclass
class SetupOptions:
    player_names: Sequence[str] = ("Player 1",)
    starting_chips: int = 1000
    decks: int = 1
    hit_soft_17: bool = True


def unreachable() -> Never:
    raise RuntimeError("pyblackjack internal error: unreachable code reached")



import argparse
import importlib
import pickle
import pickletools

from typing_extensions import assert_never, cast

from .model import Game, Hand, Player
from .utils import SAVE_DIR, SAVE_EXT, Action, StartOption
from .views._abc import AbstractBlackjackView, PlayerQuit, SaveGame


class BlackjackController:

    game: Game
    view: AbstractBlackjackView

    def __init__(self, view: AbstractBlackjackView):
        self.view = view

    def mainloop(self) -> None:
        while self.play_hand():
            pass

    def play_hand(self) -> bool:
        new_shoe = self.game.shoe.needs_shuffle()
        self.view.display_new_hand(new_shoe=new_shoe)
        if new_shoe:
            self.game.shoe.new()
        self.game.active_players.clear()
        self.collect_bets()
        if not self.game.players:
            return False
        self.deal_cards()
        self.display_hands()
        self.view.display_start_of_actions()
        if self.check_dealer_blackjack():
            return True
        self.check_player_blackjacks()
        for player in self.game.active_players.copy():
            self.play_player_hand(player)
        self.view.display_end_of_actions()
        if self.game.active_players:
            self.play_dealer_hand()
            self.pay_winners()
        else:
            self.game.dealer.flip()
            self.view.display_dealer_hand(self.game.dealer)
        return True

    def collect_bets(self) -> None:
        for player in self.game.players.copy():
            player.reset()
            self.get_bet(player)

    def get_bet(self, player: Player) -> None:
        player.hands = []
        if player.chips == 0:
            self.view.display_player_elimination(player)
            self.game.players.remove(player)
            return
        try:
            bet = self.view.get_bet(player,
                                    save=player is self.game.players[0])
        except SaveGame:
            self.save_game()
            try:
                bet = self.view.get_bet(player, save=False)
            except PlayerQuit:
                self.game.players.remove(player)
                return
        except PlayerQuit:
            self.game.players.remove(player)
            return
        player.bet = bet
        if bet > 0:
            player.chips -= bet
            self.game.active_players.append(player)
            player.hands.append(Hand(bet=bet))

    def deal_cards(self) -> None:
        self.game.dealer.reset()
        for _ in range(2):
            for hand in self.game.active_players + [self.game.dealer]:
                hand.hit()

    def display_hands(self) -> None:
        self.view.display_dealer_hand(self.game.dealer)
        for player in self.game.active_players:
            self.view.display_player_hand(player, 1)

    def check_dealer_blackjack(self) -> bool:
        if self.game.dealer.has_ace_up():
            self.handle_insurance()
        if self.game.dealer.is_blackjack():
            self.game.dealer.flip()
            self.view.display_dealer_hand(self.game.dealer)
            self.resolve_dealer_blackjack()
            return True
        for player in self.game.active_players:
            player.insurance = None
        return False

    def handle_insurance(self) -> None:
        for player in self.game.active_players:
            if (player.can_afford_insurance()
                    and self.view.get_insurance_option(player)):
                player.buy_insurance()

    def resolve_dealer_blackjack(self) -> None:
        self.view.display_dealer_blackjack(self.game.dealer)
        for player in self.game.active_players:
            if player.insurance:
                player.chips += player.insurance * 3
                player.insurance = None
            if player.has_blackjack():
                self.view.display_blackjack(player)
                player.chips += player.bet
        self.game.active_players.clear()

    def check_player_blackjacks(self) -> None:
        for player in self.game.active_players.copy():
            if player.has_blackjack():
                self.view.display_blackjack(player)
                player.chips += int(player.bet * 2.5)
                self.game.active_players.remove(player)

    def play_player_hand(self, player: Player) -> None:
        for _ in player:
            self.play_single_hand(player)
        hands_busted: list[bool] = [hand.is_busted() for _, hand in player]
        if all(hands_busted):
            self.game.active_players.remove(player)

    def play_single_hand(self, player: Player) -> None:
        hand: Hand = player.active_hand
        i: int = player.active_hand_index + 1
        if len(hand) == 1:
            # split hand
            player.hit()
        while not hand.is_busted():
            self.view.display_player_hand(player, i)
            if player.is_split() and hand[0] == 1:
                # split aces auto-stand
                break
            if hand.total() == 21:
                # 21 auto-stands
                break
            action: Action = self.get_action(player, hand)
            if not self.perform_action(action, player):
                break
        else:
            self.view.display_player_hand(player, i)
            self.view.display_bust(player, i)

    def get_action(self, player: Player, hand: Hand) -> Action:
        action_set: Action = Action.HIT | Action.STAND
        low_chips_set: Action = Action(0)
        if len(hand) == 2:
            if player.can_split():
                if player.can_afford_double():
                    action_set |= Action.SPLIT
                else:
                    low_chips_set |= Action.SPLIT
            if player.can_double():
                if player.can_afford_double():
                    action_set |= Action.DOUBLE_DOWN
                else:
                    low_chips_set |= Action.DOUBLE_DOWN
            if not player.is_split():
                action_set |= Action.SURRENDER
        return self.view.get_action(action_set, low_chips_set)

    def perform_action(self, action: Action, player: Player) -> bool:
        match action:
            case Action.HIT:
                player.hit()
                return True
            case Action.STAND:
                return False
            case Action.SPLIT:
                player.split()
                return True
            case Action.DOUBLE_DOWN:
                player.double_down()
                self.view.display_player_hand(player, player.active_hand_index + 1)
                return False
            case Action.SURRENDER:
                player.surrender()
                self.game.active_players.remove(player)
                return False
            case unreachable:
                assert_never(unreachable)

    def play_dealer_hand(self) -> None:
        dealer = self.game.dealer
        dealer.flip()
        while True:
            soft_total: int | None = dealer.soft_total()
            if ((soft_total and soft_total >= dealer.soft_stand)
                    or dealer.hard_total() >= dealer.hard_stand):
                break
            dealer.hit()
        self.view.display_dealer_hand(dealer)

    def pay_winners(self) -> None:
        for player in self.game.active_players:
            self.resolve_player_hands(player)

    def resolve_player_hands(self, player: Player) -> None:
        hand: Hand
        for index, hand in player:
            if hand.is_busted():
                continue
            if (self.game.dealer.is_busted()
                    or hand.total() > self.game.dealer.total()):
                player.chips += hand.bet * 2
                self.view.display_win(player, index)
            elif hand.total() == self.game.dealer.total():
                player.chips += hand.bet
                self.view.display_push(player, index)
            else:
                self.view.display_loss(player, index)

    def save_game(self, filename: str | None = None, /) -> None:
        if not SAVE_DIR.exists():
            SAVE_DIR.mkdir()
        if filename is None:
            filename = self.view.get_save_filename()
        path = SAVE_DIR / f"{filename}{SAVE_EXT}"
        if path.exists():
            if not self.view.confirm_overwrite():
                return
        pkl: bytes = pickle.dumps(self.game)
        pkl = pickletools.optimize(pkl)
        path.write_bytes(pkl)

    def load_game(self) -> None:
        filename = self.view.get_load_game()
        if filename is None:
            self.game = Game(self.view.get_setup_options())
            return
        else:
            with (SAVE_DIR / filename).open("rb") as file:
                self.game = cast(Game, pickle.load(file))

    def main(self) -> int:
        match self.view.get_start_option():
            case StartOption.NEW_GAME:
                self.game = Game(self.view.get_setup_options())
            case StartOption.LOAD_GAME:
                self.load_game()
            case unreachable:
                assert_never(unreachable)
        # For debugging: stack cards to top to reproduce bugs
        #                            D   P1B  P1A  P1A  Dup  P1B  Ddn  P1A
        #stacked_cards: list[str] = ["9", "A", "J", "5", "K", "8", "6", "8"]
        #from .model import Card
        #for rank in stacked_cards:
        #    self.game.shoe.cards.append(Card(rank))
        self.mainloop()
        self.view.display_exit()
        return 0

def entrypoint() -> int:
    parser = argparse.ArgumentParser()
    view_group = parser.add_mutually_exclusive_group(required=False)
    view_group.add_argument("--boss",
                            const="boss",
                            action="store_const",
                            dest="view")
    view_group.add_argument("-v", "--view", action="store", dest="view")
    parser.set_defaults(view="terminal")
    args = parser.parse_args()
    view_mod = importlib.import_module(f"pyblackjack.views.{args.view}")
    controller = BlackjackController(view_mod.BlackjackView())
    return controller.main()



import sys

from typing_extensions import override

from . import boss
from ..utils import SAVE_DIR


class BlackjackView(boss.BlackjackView):

    def __init__(self) -> None:
        super().__init__()
        # pylint: disable-next=consider-using-with
        sys.stderr = open(SAVE_DIR / "stderr.log", "w", encoding="utf-8")
        self._print_codes(
            self.term.enter_fullscreen,
            self.term.home,
            self.term.black_on_green,
            self.term.clear,
        )

    def __del__(self) -> None:
        self._print_codes(self.term.exit_fullscreen)
        sys.stderr.close()

    @override
    def display_exit(self) -> None:
        super().display_exit()
        print("Press any key to exit...")
        with self.term.cbreak():
            self.term.inkey()

    def _print_codes(self, *codes: str) -> None:
        for code in codes:
            print(code, end="", flush=True)



from __future__ import annotations

from collections import Counter
import random

from typing_extensions import Any, ClassVar, Final, Iterator, cast, no_type_check, override

from .utils import SetupOptions, unreachable


class Card(int):
    '''An individual playing card. Subclass of int for simplicity.'''

    rank: str

    def __new__(cls, rank: str) -> Card:
        value: int
        if rank == 'A':
            value = 1
        elif rank in 'JQK':
            value = 10
        else:
            value = int(rank)
        self: Card = super().__new__(cls, value)
        self.rank = rank
        return self

    def __str__(self) -> str:
        return str(self.rank)


CARDSET = [
    Card('A'),
    Card('2'),
    Card('3'),
    Card('4'),
    Card('5'),
    Card('6'),
    Card('7'),
    Card('8'),
    Card('9'),
    Card('10'),
    Card('J'),
    Card('Q'),
    Card('K'),
]

DECK: Final[list[Card]] = CARDSET * 4


class Shoe:

    def __init__(self, *, decks: int = 6):
        self.decks: int = decks
        self.cards: list[Card] = []
        self.cut: int = 0
        self.new()

    def deal(self) -> Card:
        return self.cards.pop()

    def needs_shuffle(self) -> bool:
        return len(self.cards) <= self.cut

    def new(self) -> None:
        self.cards = DECK * self.decks
        random.shuffle(self.cards)
        if self.decks == 1:
            self.cut = len(self.cards) // 2
        else:
            self.cut = len(self.cards) // 4
        self.cut += random.randint(self.decks * -4, self.decks * 4)

    def __getstate__(self) -> tuple[int, dict[str, int], int, int]:
        version = 1
        counter = Counter(card.rank for card in self.cards)
        return (version, dict(counter), self.decks, self.cut)

    def __setstate__(self, state: tuple[Any, ...]) -> None:
        match state:
            case (1, dict() as counter, int() as decks, int() as cut):
                self.decks = decks
                self.cut = cut
                self.cards = []
                for rank, count in cast(dict[str, int], counter).items():
                    self.cards.extend([Card(rank)] * count)
                random.shuffle(self.cards)
            case _:
                unreachable()


class Hand(list[Card]):

    def __init__(self, *, bet: int) -> None:
        super().__init__()
        self.bet: int = bet

    def hard_total(self) -> int:
        return sum(self)

    def soft_total(self) -> int | None:
        total = self.hard_total()
        if total > 11 or 1 not in self:
            return None
        return total + 10

    def total(self) -> int:
        return self.soft_total() or self.hard_total()

    def is_blackjack(self) -> bool:
        return len(self) == 2 and self.total() == 21

    def is_busted(self) -> bool:
        return self.hard_total() > 21

    def hit_from_shoe(self, shoe: Shoe) -> None:
        self.append(shoe.deal())

    def reset(self) -> None:
        self.clear()


class Dealer(Hand):

    def __init__(self, shoe: Shoe, hit_soft_17: bool = True):
        super().__init__(bet=0)
        self.hard_stand: Final[int] = 17
        self.soft_stand: Final[int] = 18 if hit_soft_17 else 17
        self.downcard: bool = True
        self.shoe: Shoe = shoe

    def has_ace_up(self) -> bool:
        return self[1] == 1

    def hit(self) -> None:
        self.hit_from_shoe(self.shoe)

    def flip(self) -> None:
        assert self.downcard is True
        self.downcard = False

    @override
    def reset(self) -> None:
        super().reset()
        self.downcard = True

    def __getstate__(self) -> tuple[int, int]:
        super().reset()
        version = 1
        return (version, self.soft_stand)

    @no_type_check
    def __setstate__(self, state: tuple[Any, ...]) -> None:
        super().__init__(bet=0)
        match state:
            case (1, int() as soft_stand):
                self.soft_stand = soft_stand
            case _:
                unreachable()
        self.hard_stand = 17


class Player:
    _id: ClassVar[int] = 0

    def __init__(
        self,
        shoe: Shoe,
        name: str,
        chips: int = 1000,
    ):
        self.__class__._id += 1
        self.name: Final[str] = name
        self.chips: int = chips
        self.bet: int = 0
        self.insurance: int | None = None
        self.hands: list[Hand] = []
        self.active_hand_index: int = 0
        self.shoe: Shoe = shoe

    def __iter__(self) -> Iterator[tuple[int, Hand]]:
        while self.active_hand_index < len(self.hands):
            yield (self.active_hand_index + 1, self.active_hand)
            self.active_hand_index += 1
        # cleanup
        self.active_hand_index = 0

    @property
    def active_hand(self) -> Hand:
        return self.hands[self.active_hand_index]

    def is_split(self) -> bool:
        return len(self.hands) > 1

    def can_afford_double(self) -> bool:
        return self.chips >= self.bet

    def has_blackjack(self) -> bool:
        return not self.is_split() and self.active_hand.is_blackjack()

    def hit(self) -> None:
        self.active_hand.hit_from_shoe(self.shoe)

    def can_split(self) -> bool:
        hand: Hand = self.active_hand
        return len(hand) == 2 and hand[0].rank == hand[1].rank

    def split(self) -> None:
        assert self.can_afford_double()
        assert self.can_split()
        self.chips -= self.bet
        new_hand = Hand(bet=self.bet)
        new_hand.append(self.active_hand.pop())
        self.hands.append(new_hand)
        self.active_hand.hit_from_shoe(self.shoe)

    def can_double(self) -> bool:
        return (len(self.active_hand) == 2
                and 9 <= self.active_hand.total() <= 11)

    def double_down(self) -> None:
        assert self.can_afford_double()
        assert len(self.active_hand) == 2
        self.chips -= self.active_hand.bet
        self.active_hand.bet *= 2
        self.active_hand.hit_from_shoe(self.shoe)

    def surrender(self) -> None:
        assert len(self.active_hand) == 2
        self.chips += self.bet // 2

    def can_afford_insurance(self) -> bool:
        return self.chips >= self.bet // 2

    def buy_insurance(self) -> None:
        assert self.can_afford_insurance()
        self.insurance = self.bet // 2
        self.chips -= self.insurance

    def reset(self) -> None:
        self.hands.clear()
        self.active_hand_index = 0
        self.insurance = None

    def __getstate__(self) -> tuple[int, str, int]:
        version = 1
        return (version, self.name, self.chips)

    @no_type_check
    def __setstate__(self, state: tuple[Any, ...]) -> None:
        match state:
            case (1, str() as name, int() as chips):
                self.name = name
                self.chips = chips
            case _:
                unreachable()
        self.bet = 0
        self.insurance = None
        self.hands = []
        self.active_hand_index = 0


class Game:

    def __init__(self, options: SetupOptions):
        self.shoe: Final[Shoe] = Shoe(decks=options.decks)
        self.dealer: Final[Dealer] = Dealer(self.shoe, hit_soft_17=options.hit_soft_17)
        self.players: Final[list[Player]] = [
            Player(self.shoe, name, options.starting_chips)
            for name in options.player_names
        ]
        self.active_players: list[Player] = []

    def activate_player(self, player: Player) -> None:
        self.active_players.append(player)

    def deactivate_player(self, player: Player) -> None:
        self.active_players.remove(player)

    def __getstate__(self) -> tuple[int, Dealer, list[Player], Shoe]:
        version = 1
        return (version, self.dealer, self.players, self.shoe)

    @no_type_check
    def __setstate__(self, state: tuple[Any, ...]) -> None:
        match state:
            case (1, Dealer() as dealer, list() as players, Shoe() as shoe):
                # Dealer, list[Player], Shoe
                self.dealer = dealer
                self.players = cast(list[Player], players)
                self.shoe = shoe
            case _:
                unreachable()
        self.dealer.shoe = self.shoe
        for player in self.players:
            player.shoe = self.shoe
        self.active_players = []
